#!/usr/bin/python3

# ###############################
#
# timecalc 0.7
# Code made by bgo@vedur.is
# Iceland Met Office
# 2015
#
# ###############################

"""
#!/usr/bin/python

"""
import gtimes.timefunc as timefunc

import argparse
import datetime
import sys
import re


def datestr(string):
    """
    Formating check for -d option
    """
    if type(string) != datetime.datetime:  # more to contitions to come
        msg = "%r is not correctly formated" % string
        raise argparse.ArgumentTypeError(msg)
    return string


def main():
    """
    command line script for
    general time and date calculations.
    can handle obscure GPS time formating.

    run: timecalc -h for help
    """

    # date to use defaults to today ----
    dstr = "%Y-%m-%d"  # Default input string
    outpstr = "%a, %d. %b %Y"  # Default output string

    # today=datetime.date.today().strftime(dstr)
    today = datetime.datetime.utcnow()

    dD_parser = argparse.ArgumentParser(
        description="Time conversion program", add_help=False
    )
    dD_parser.add_argument(
        "-D",
        default=0,
        nargs="?",
        const=1,
        type=float,
        help="Number of days to shift the given day positive subracts, negativ adds",
    )
    dD_parser.add_argument(
        "--shift",
        nargs="?",
        type=str,
        help="General shift string given for more general shift "
        + "general format is given by format letters"
        + "d for days, S for seconds,f for microseconds,"
        + "m for milliseconds, M for minutes, H for hours,"
        + "w for weeks.  with the shift as a number behind each letter"
        + "using : to seperate the different fields , positive subracts,"
        + "negativ adds. Example: 'd1:S-3:H3'",
    )
    dD_parser.add_argument(
        "-d",
        default=today,
        help="Input date. The default format is "
        "%%Y-%%m-%%m"
        "." + " The format can be modifed through the -f flag",
    )
    dD_parser.add_argument(
        "-f",
        default=dstr,
        type=str,
        help="Format of the string passed to -d. If absent, -d defaults to "
        "%%Y-%%m-%%-%%m"
        "." + " Special formating: " + " "
        "yearf"
        " -> fractional year " + " "
        "w-dow"
        " -> GPS Week-Day of Week. " + " "
        "w-sow"
        " -> GPS Week-Second of Week. " + " "
        "w-dow-sod"
        " -> GPS Week-Day of week-Second of day. "
        + "See datetime documentation for general formating",
    )

    # parse what has -d, -f and -D to use as input for other options
    args, remaining_argv = dD_parser.parse_known_args()
    dstr = args.f

    # dealing with the input format.
    day = dday = timefunc.toDatetime(args.d, args.f)

    if args.shift:
        Shift = args.shift
    else:
        Shift = args.D

    day = day - datetime.timedelta(
        **timefunc.shifTime(Shift)
    )  # applying the Day offset
    # ----------------------------------------

    parser = argparse.ArgumentParser(
        parents=[dD_parser],
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Year flag -------------------
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-o",
        type=str,
        help=" Output format for general formatting defaults to "
        "%%a, %%d. %%b %%Y"
        "."
        + " See datetime documentation for the formating."
        + " for special formatting see other arguments",
    )
    group.add_argument(
        "-l",
        "--list",
        nargs="*",
        action="store",
        type=str,
        help="Return a list of strings according to the format string passed"
        + "Special cases are: #b -> three letter lower case month %%b.lower() i.e. jan, "
        + "#gpsw -> gps week number in  wwww, "
        + "#Rin2 -> rinex 2 format %%j(session).%%y, "
        + "where session is a single character session identifier, "
        + "#8hRin2 -> special case of 8h rinex two files where Sessions are {0,1,2}) "
        + "Example:  /data/%%Y/#b/VONC/15s_24hr/rinex/VONCRin2O.Z -> "
        + "/data/2015/oct/VONC/15s_24hr/rinex/VONC2740.15O.Z \n"
        + "#datelist -> Specal case. list of dates one in a line. "
        + "Use -f to change the input format",
    )
    group.add_argument(
        "-p",
        type=str,
        help="Second input date, will calculate the number of days between the"
        + "input date from -d (or today) and the date given with -p."
        + "The default format is "
        "%%Y-%%m-%%m"
        "." + " The format can be modifed through the -f flag",
    )
    group.add_argument(
        "-y",
        type=int,
        nargs="?",
        const=timefunc.shlyear(yyyy=day.year, change=True),
        help="Returns the Year in "
        "YYYY"
        " form. Special case: If the year is passed directly"
        + "to -y it is converted to two/four digit form, depending on the input.",
    )
    group.add_argument(
        "-yy",
        action="store_true",
        help="Returns the Year from -d in two digit form ("
        "YY"
        "). " + "Current year is returned in two digit form if -d is omitted",
    )
    group.add_argument(
        "-yf",
        action="store_const",
        const=day,
        help="Return date from -d as fractional year. "
        + "Current date is returned if -d omitted",
    )
    group.add_argument(
        "-H",
        action="store_const",
        const=day,
        help="Return date from -d as an Hour of day ",
    )
    # flags for different output
    group.add_argument(
        "-t",
        action="store_const",
        const=day,
        help="Return a space seperated string of "
        + " (year, month, day of month, day of year, fractional year, GPS week,"
        + "day of week sunday = 0 ... 6 = saturday) -> "
        + "(YYYY MM DD DOY YYYY.ddd WWWW DOW) ",
    )
    group.add_argument(
        "-j", action="store_const", const=day, help="Return the day of year"
    )
    group.add_argument(
        "-ny",
        action="store_const",
        const=day,
        help="Return the number of days in thye year",
    )
    group.add_argument(
        "-w", "--week", action="store_const", const=day, help="Return GPS week"
    )
    group.add_argument(
        "-u",
        "--dayofw",
        action="store_const",
        const=day,
        help="Return day of GPS week sunday = 0 ... 6 = saturday ",
    )
    group.add_argument(
        "-wd",
        "--weekday",
        action="store_const",
        const=day,
        help="Return GPS week and day of GPS week sunday = 0 ... 6 = saturday ",
    )
    group.add_argument(
        "-ws",
        "--wsow",
        action="store_const",
        const=day,
        help="Return GPS week and Second of GPS week",
    )
    group.add_argument(
        "-i",
        "--datet",
        action="store_const",
        const=day,
        help="Returns the calendar date and time",
    )
    group.add_argument(
        "-r",
        "--rinex",
        const="d",
        nargs="?",
        choices=["H", "d"],
        help="return time part rinex standard hourly format format",
    )
    group.add_argument(
        "--GPST",
        const="all",
        nargs="?",
        choices=[
            "all",
            "w",
            "wdow",
            "dow",
            "wsow",
            "sow",
            "wdowsod",
            "dowsod",
            "dow",
            "sod",
        ],
        help="Return a different compinations of GPS time, GPS week (w), Second of week (sow)"
        + "Day of week (dow), Second of day (sod)",
    )
    args = parser.parse_args(args=remaining_argv)
    # ------------------------------

    # printing out stuff depending on args
    if args.y != None:
        print(timefunc.shlyear(yyyy=args.y, change=True))

    elif args.ny:
        print(timefunc.DaysinYear(day.year))

    elif args.yy:
        print(timefunc.shlyear(yyyy=day.year, change=True))

    elif args.yf:
        print(timefunc.currYearfDate(refday=day))

    elif args.H:
        print(day.hour)

    elif args.j:
        print(timefunc.currDate(refday=args.j, String="%j"))

    elif args.week:
        print(timefunc.gpsWeekDay(refday=args.week)[0])

    elif args.dayofw:
        print(timefunc.gpsWeekDay(refday=args.dayofw)[1])

    elif args.weekday:
        print("%04d %03d" % timefunc.gpsWeekDay(refday=args.weekday))

    elif args.GPST:
        (GPSW, SOW, DOW, SOD) = timefunc.gpsfDateTime(refday=day)

        if args.GPST == "all":
            print("{0:d} {1:.0f} {2:03d} {3:.0f}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "w":
            print("{0:d}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "wdow":
            print("{0:d} {2:03d}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "dow":
            print("{2:03d}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "wsow":
            print("{0:d} {1:.0f}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "sow":
            print("{1:.0f}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "wdowsod":
            print("{0:d} {2:03d} {3:.0f}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "dowsod":
            print("{2:03d} {3:.0f}".format(GPSW, SOW, DOW, SOD))
        elif args.GPST == "sod":
            print("{3:.0f}".format(GPSW, SOW, DOW, SOD))

    elif args.wsow:
        print("{0:d}-{1:.0f}".format(*timefunc.gpsfDateTime(refday=day)[0:2]))

    elif args.datet:
        print(day.isoformat())

    elif args.t:
        print(
            "%d %02d %02d %03d %4.5f %04d %d %d %02d %s"
            % timefunc.dateTuple(refday=args.t)
        )

    elif args.o:
        print(day.strftime(args.o))

    elif args.list:
        # error check for argument --list
        if len(args.list) not in (2, 3):
            parser.error("Either give two or three arguments with --list")
        else:
            if len(args.list) == 2:  # the default for list[2]
                args.list.append("right")

            if args.list[2] not in {"left", "right"}:
                args.list[2] = None

            stringlist = timefunc.datepathlist(
                args.list[0], args.list[1], day, dday, closed=args.list[2]
            )

            datelistmatch = re.compile(r"\w*(#datelist)\w*").search(args.list[0])
            if datelistmatch:
                stringlist = [i.strftime("{0}\n".format(dstr)) for i in stringlist]

            [sys.stdout.write(item) for item in stringlist]
            sys.stdout.write("\n")  # to start with a newline after printing the list

    elif args.p:
        # dealing with the input format.
        day1 = timefunc.toDatetime(args.p, dstr)
        if dstr == "yearf":
            print((day - day1))
        else:
            print((day - day1).days)

    elif args.rinex:
        if args.rinex == "H":
            hour = timefunc.hourABC(day.hour)
        elif args.rinex == "d":
            hour = 0

        doy = int(day.strftime("%j"))
        yr = int(day.strftime("%y"))
        print("%.3d%s.%.2d" % (doy, hour, yr))

    else:
        print(day.strftime(outpstr))


if __name__ == "__main__":
    main()



"""
A Python implementation of GPS related time conversions.

Copyright 2002 by Bud P. Bruegger, Sistema, Italy
mailto:bud@sistema.it
http://www.sistema.it

Modifications for GPS seconds by Duncan Brown

PyUTCFromGpsSeconds added by Ben Johnson

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place, Suite 330, Boston, MA  02111-1307  USA

GPS Time Utility functions

This file contains a Python implementation of GPS related time conversions.

The two main functions convert between UTC and GPS time (GPS-week, time of
week in seconds, GPS-day, time of day in seconds).  The other functions are
convenience wrappers around these base functions.

A good reference for GPS time issues is:
http://www.oc.nps.navy.mil/~jclynch/timsys.html

Note that python time types are represented in seconds since (a platform
dependent Python) Epoch.  This makes implementation quite straight forward
as compared to some algorigthms found in the literature and on the web.
"""

__author__ = "Benedikt G. Ofeigsson <bgo@vedur.is> edited from Duncan Brown <duncan@gravity.phys.uwm.edu>"
__date__ = "$Date: 2012/03/6"
__version__ = "$Revision: 1.6 $"[11:-2]
# $Source: /usr/local/cvs/lscsoft/glue/glue/gpstime.py,v $

import time
import math
import datetime

secsInWeek = 604800
secsInDay = 86400
gpsEpoch = (1980, 1, 6, 0, 0, 0)  # (year, month, day, hh, mm, ss)
epochTuple = gpsEpoch + (-1, -1, 0)


def dayOfWeek(year, month, day):

    """
    Returns day of week: 0=Sun, 1=Mon, .., 6=Sat
    """
    hr = 12  # make sure you fall into right day, middle is save
    t = time.mktime((year, month, day, hr, 0, 0, 0, 0, -1))
    pyDow = time.localtime(t)[6]
    gpsDow = (pyDow + 1) % 7
    return gpsDow


def gpsWeek(year, month, day):
    """
    Returns (full) gpsWeek for given date (in UTC)
    """
    hr = 12  # make sure you fall into right day, middle is save
    return gpsFromUTC(year, month, day, hr, 0, 0)[0]


def julianDay(year, month, day):
    "returns julian day=day since Jan 1 of year"
    hr = 12  # make sure you fall into right day, middle is save
    t = time.mktime((year, month, day, hr, 0, 0.0, 0, 0, -1))
    julDay = time.localtime(t)[7]
    return julDay


def mkUTC(year, month, day, hour, min, sec):
    """
    Similar to python's mktime but for utc
    """
    spec = [year, month, day, hour, min, sec] + [0, 0, 0]
    utc = time.mktime(spec) - time.timezone
    return utc


def ymdhmsFromPyUTC(pyUTC):
    """
    Returns tuple from a python time value in UTC
    """

    return time.gmtime(pyUTC)[:-3]


# def wtFromUTCpy(pyUTC, leapSecs=18):
#    """
#    Convenience function:
#    BGO 07.01.2013: changed the default leapsecond to 16
#    BGO 07.01.2019: changed the default leapsecond to 18
#         allows to use python UTC times and
#         returns only week and tow"""
#
#    ymdhms = ymdhmsFromPyUTC(pyUTC)
#    wSowDSoD = gpsFromUTC( *ymdhms)
#    return wSowDSoD[0:2]
#    #return ymdhms


def gpsFromUTC(year, month, day, hour, min, sec, leapSecs=None, gpst=True):
    """
    Converts UTC to: gpsWeek, secsOfWeek, gpsDay, secsOfDay

    a good reference is:  http://www.oc.nps.navy.mil/~jclynch/timsys.html

    This is based on the following facts (see reference above):

    GPS time is basically measured in (atomic) seconds since
    January 6, 1980, 00:00:00.0  (the GPS Epoch)

    The GPS week starts on Saturday midnight (Sunday morning), and runs
    for 604800 seconds.

    Currently, GPS time is 13 seconds ahead of UTC (see above reference).
    While GPS SVs transmit this difference and the date when another leap
    second takes effect, the use of leap seconds cannot be predicted.  This
    routine is precise until the next leap second is introduced and has to be
    updated after that.
    BGO 07.01.2013: changed the default leapsecond to 16
    BGO 07.01.2019: changed the default leapsecond to 18

    SOW = Seconds of Week
    SOD = Seconds of Day

    Note:  Python represents time in integer seconds, fractions are lost!!!
    """
    import datetime as dt

    # Automatic or manually applyed leap seconds.
    if leapSecs is None:
        leapSecs = getleapSecs(
            dTime=dt.datetime(year, month, day, hour, min, sec), gpst=gpst
        )

    secFract = sec % 1
    t0 = time.mktime(epochTuple)
    t = time.mktime((year, month, day, hour, min, sec, -1, -1, 0))
    # Note: time.mktime strictly works in localtime and to yield UTC, it should be
    #       corrected with time.timezone
    #       However, since we use the difference, this correction is unnecessary.
    # Warning:  trouble if daylight savings flag is set to -1 or 1 !!!
    t = t + leapSecs
    tdiff = t - t0
    gpsSOW = (tdiff % secsInWeek) + secFract
    gpsWeek = int(math.floor(tdiff / secsInWeek))
    gpsDay = int(math.floor(gpsSOW / secsInDay))
    gpsSOD = gpsSOW % secsInDay

    return (gpsWeek, gpsSOW, gpsDay, gpsSOD)


def UTCFromGps(gpsWeek, SOW, leapSecs=None, dtimeObj=False):
    """
    Converts gps week and seconds to UTC
    BGO 07.01.2013: changed the default leapsecond to 16
    BGO 07.01.2019: changed the default leapsecond to 18

    see comments of inverse function!

    SOW = seconds of week
    gpsWeek is the full number (not modulo 1024)
    """
    import datetime as dt

    # Automatic or manually applyed leap seconds.
    if leapSecs is None:
        leapSecs = getleapSecs(dTime=(gpsWeek, SOW), gpst=True)

    secFract = SOW % 1
    epochTuple = gpsEpoch + (-1, -1, 0)
    t0 = time.mktime(epochTuple) - time.timezone  # mktime is localtime, correct for UTC
    tdiff = (gpsWeek * secsInWeek) + SOW - leapSecs
    t = t0 + tdiff
    if dtimeObj is True:
        # only deals with whole seconds
        return dt.datetime.fromtimestamp(t)
    else:
        (
            year,
            month,
            day,
            hh,
            mm,
            ss,
            dayOfWeek,
            julianDay,
            daylightsaving,
        ) = time.gmtime(t + secFract)
        # use gmtime since localtime does not allow to switch off daylighsavings correction!!!
        return (year, month, day, hh, mm, ss + secFract)


def GpsSecondsFromPyUTC(pyUTC, leapSecs=None):
    """
    Converts the python epoch to gps seconds
    BGO 07.01.2013: changed the default leapsecond to 16
    BGO 07.01.2019: changed the default leapsecond to 18

    pyEpoch = the python epoch from time.time()
    """
    import datetime as dt

    # Automatic or manually applyed leap seconds.
    if leapSecs is None:
        leapSecs = getleapSecs(dTime=dt.datetime(*ymdhmsFromPyUTC(pyUTC)), gpst=True)

    t = gpsFromUTC(*ymdhmsFromPyUTC(pyUTC - leapSecs))

    return int(t[0] * 60 * 60 * 24 * 7 + t[1])


def PyUTCFromGpsSeconds(gpsseconds):
    """
    Converts gps seconds to the
    python epoch. That is, the time
    that would be returned from time.time()
    at gpsseconds.
    """

    pyUTC


def getleapSecs(dTime=None, gpst=True):
    """
    Retreiving the leap second for a given datetime
    Input:
        dTime: a Time string can be a datetime object, string or integer (either seconds since Epoch or GPSs)
    """

    import time
    import datetime as dt
    import locale

    # tdiff = (gpsWeek * secsInWeek) + SOW - leapSecs
    # t = t0 + tdiff

    leapsecdict = leapSecDict()
    leapSecs = 0

    if type(dTime) is datetime.datetime:
        pass
    elif (type(dTime) is int) or (type(dTime) is float):
        dTime = dt.datetime.fromtimestamp(dTime)
    elif type(dTime) is tuple:
        t0 = (
            time.mktime(epochTuple) - time.timezone
        )  # mktime is localtime, correct for UTC
        tdiff = dTime[0] * secsInWeek + dTime[1]
        dTime = dt.datetime.fromtimestamp(t0 + tdiff)
    elif dTime is None:
        dTime = dt.datetime.now()

    # print(dTime.strftime("%Y-%b-%d %H:%M:%S"))
    # print(time.mktime( dTime.timetuple()) )

    # making sure locale month format is correct by changing locale.
    syslocale = locale.getlocale()
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")

    for day in leapsecdict.keys():
        if dt.datetime.strptime(day, "%Y-%b-%d") <= dTime:
            leapSecs = leapsecdict[day]

    # changing back to system local.
    locale.setlocale(locale.LC_ALL, "{}.{}".format(*syslocale))

    if gpst is True:

        leapSecs = leapSecs - leapsecdict["1980-Jan-1"]
        # print( "ADDING: {0} s".format(leapSecs) )

    return leapSecs


def leapSecDict():

    leapSecDict = {
        "1972-Jan-1": 10,
        "1972-Jul-1": 11,
        "1973-Jan-1": 12,
        "1974-Jan-1": 13,
        "1975-Jan-1": 14,
        "1976-Jan-1": 15,
        "1977-Jan-1": 16,
        "1978-Jan-1": 17,
        "1979-Jan-1": 18,
        "1980-Jan-1": 19,
        "1981-Jul-1": 20,
        "1982-Jul-1": 21,
        "1983-Jul-1": 22,
        "1985-Jul-1": 23,
        "1988-Jan-1": 24,
        "1990-Jan-1": 25,
        "1991-Jan-1": 26,
        "1992-Jul-1": 27,
        "1993-Jul-1": 28,
        "1994-Jul-1": 29,
        "1996-Jan-1": 30,
        "1997-Jul-1": 31,
        "1999-Jan-1": 32,
        "2006-Jan-1": 33,
        "2009-Jan-1": 34,
        "2012-Jul-1": 35,
        "2015-Jul-1": 36,
        "2017-Jan-1": 37,
    }

    return leapSecDict


# def gpsTimeFromUTC()

# ===== Tests  =========================================


def testTimeStuff():
    print("-" * 20)
    print("The GPS Epoch when everything began (1980, 1, 6, 0, 0, 0, leapSecs=0)")
    (w, sow, d, sod) = gpsFromUTC(1980, 1, 6, 0, 0, 0, leapSecs=0)
    print("**** week: %s, sow: %s, day: %s, sod: %s" % (w, sow, d, sod))
    print("     and hopefully back:")
    print("**** %s, %s, %s, %s, %s, %s\n" % UTCFromGps(w, sow, leapSecs=0))

    print("The time of first Rollover of GPS week (1999, 8, 21, 23, 59, 47)")
    (w, sow, d, sod) = gpsFromUTC(1999, 8, 21, 23, 59, 47)
    print("**** week: %s, sow: %s, day: %s, sod: %s" % (w, sow, d, sod))
    print("     and hopefully back:")
    print("**** %s, %s, %s, %s, %s, %s\n" % UTCFromGps(w, sow, leapSecs=14))

    print("Today is GPS week 1186, day 3, seems to run ok (2002, 10, 2, 12, 6, 13.56)")
    (w, sow, d, sod) = gpsFromUTC(2002, 10, 2, 12, 6, 13.56)
    print("**** week: %s, sow: %s, day: %s, sod: %s" % (w, sow, d, sod))
    print("     and hopefully back:")
    print("**** %s, %s, %s, %s, %s, %s\n" % UTCFromGps(w, sow))


def testJulD():
    print("2002, 10, 11 -> 284  ==??== ", julianDay(2002, 10, 11))


def testGpsWeek():
    print("2002, 10, 11 -> 1187  ==??== ", gpsWeek(2002, 10, 11))


def testDayOfWeek():
    print("2002, 10, 12 -> 6  ==??== ", dayOfWeek(2002, 10, 12))
    print("2002, 10, 6  -> 0  ==??== ", dayOfWeek(2002, 10, 6))


def testPyUtilties():
    ymdhms = (2002, 10, 12, 8, 34, 12.3)
    print("testing for: ", ymdhms)
    pyUtc = apply(mkUTC, ymdhms)
    back = ymdhmsFromPyUTC(pyUtc)
    print("yields     : ", back)
    # *********************** !!!!!!!!
    # assert(ymdhms == back)
    # TODO:: this works only with int seconds!!! fix!!!
    (w, t) = wtFromUTCpy(pyUtc)
    print("week and time: ", (w, t))


# ===== Main =========================================
if __name__ == "__main__":
    pass
    testTimeStuff()
    testGpsWeek()
    testJulD()
    testDayOfWeek()
    testPyUtilties()



#!/usr/bin/python3

# ###############################
#
# timefunc.py 0.7
# Code made by bgo@vedur.is
# Iceland Met Office
# 2015
#
# ###############################

"""

In this program are the following functions in order:


TimetoYearf(year, month, day, hour=12, minute=0, sec=0):
TimefromYearf(yearf, string=None):
currDatetime(days=0,refday=datetime.datetime.today(),string=None):
currDate(days=0,refday=datetime.datetime.today(),string=None,fromYearf=False):
gpsWeekDay( days=0, refday=currDate() , fromYearf=False):
currTime(string):
DayofYear(days=0, year=None, month=None, day=None):
DaysinYear( year=None ):
yearDoy(yearf):
currYearfDate( days=0, refday=datetime.date.today(), fromYearf=True ):
currYear():
shlyear(yyyy=currYear(),change=True):
dateTuple(days=0,refday=datetime.datetime.today(),string=None,fromYearf=False):
datefgpsWeekDOW(gpsWeek,DOW,string=None,leapSecs=18):
datefgpsWeekDoy(gpsWeek,Doy,string=None,leapSecs=18):
toDatetime(dStr,fStr):
_to_ordinalf(dt):
shlyear(yyyy=currYear(),change=True):
d
"""


import math
import calendar
import datetime
import string


# importing constants from gpstime.
from gtimes.gpstime import secsInDay, gpsFromUTC, UTCFromGps
from dateutil.tz import tzlocal


# Core functions ---------------------------
def shifTime(String="d0"):
    """ """

    Unitdict = {
        "d": "days",
        "S": "seconds",
        "f": "microseconds",
        "m": "milliseconds",
        "M": "minutes",
        "H": "hours",
        "w": "weeks",
    }

    Shiftdict = {
        "days": 0.0,
        "seconds": 0.0,
        "microseconds": 0.0,
        "milliseconds": 0.0,
        "minutes": 0.0,
        "hours": 0.0,
        "weeks": 0.0,
    }

    if type(String) is not str:
        String = "d" + str(String)

    for i in String.split(":"):
        Shiftdict[Unitdict[i[0]]] = float(i[1:])

    return Shiftdict


def TimetoYearf(year, month, day, hour=12, minute=0, sec=0):
    """
    converts time cal+time of day into fractional year.
    input:
        date: as year, month, day
        time: as hour, minute, sec (defaults to the middle of the day at 12 am)
    output:
        returns fractional year

    """
    doy = DayofYear(0, year, month, day) - 1
    secofyear = doy * secsInDay + (hour * 60 + minute) * 60 + sec

    daysinyear = DaysinYear(year)
    secinyear = daysinyear * secsInDay

    yearf = year + secofyear / float(secinyear)

    return yearf


def TimefromYearf(yearf, String=None):
    """
    Returns a date and/or time according to a formated string derived from a fractional year.
    Intended to manipulate time format of gamit time series files which is in fractional years.
    Input:
        string:
            formated according to format codes that the C standard (1989 version) see documentation for
            datetime module. Example "%Y %m %d %H:%M:%S %f"
        year: fractional year. example 2012.55

    Output: Returns time of the input year formated according to input string.
    """
    # to integer year
    year = int(math.floor(yearf))

    # converting to doy, hour, min, sec, microsec
    daysinyear = DaysinYear(year)
    dayf = (yearf - year) * daysinyear + 1
    doy = int(math.floor(dayf))  # day of year)
    fofday = dayf - doy
    Hour = int(math.floor((fofday) * 24))  # hour of day
    Min = int(math.floor((fofday) * 24 * 60 % 60))  # minute of hour
    fsec = fofday * 24 * 60 * 60 % 60
    Sec = int(math.floor(fsec))  # second of minute
    musec = int(math.floor((fsec - Sec) * 1000000))  # microsecond 0 - 1000000

    timestr = "%d %.3d %.2d:%.2d:%.2d %s" % (year, doy, Hour, Min, Sec, musec)
    dt = datetime.datetime.strptime(
        timestr, "%Y %j %H:%M:%S %f"
    )  # Create datetime object from timestr
    if String:
        if String == "ordinalf":  # return a floating point ordinal day
            return dt.toordinal() + fofday
        else:
            return dt.strftime(String)
    else:  # just return the datetime instanse
        return dt


def currDatetime(days=0, refday=datetime.datetime.today(), String=None):
    """
    Returns a datetime object for the date, "days" from refday.

    Input:
        days: integer, Defaults to 0
              days to offset
        refday: datetime object or a string, defaults to datetime.datetime.today()
              reference day
        string: formating string. defaults to None (infering refday as datetime object)
              If refday is a date string, this has to contain it's formating (i.e %Y-%m-%d %H:%M)

    Output:
        returns a datetime object

         defaults to current day if ran without arguments
    """

    day = refday + datetime.timedelta(**shifTime(days))
    if String:
        return day.strftime(String)
    else:
        return day


def currDate(days=0, refday=datetime.date.today(), String=None, fromYearf=False):
    """
    Returns a datetime object for the date, "days" from refday.
    Defaults to current day
    """

    if fromYearf and type(refday) == float or type(refday) == int:
        refday = TimefromYearf(refday)

    day = refday + datetime.timedelta(**shifTime(days))
    if String == "yearf":
        return TimetoYearf(*day.timetuple()[0:3])
    elif String:
        return day.strftime(String)
    else:
        return day


def gpsfDateTime(
    days=0, refday=currDatetime(), fromYearf=False, mday=False, leapSecs=None, gpst=True
):
    """
    Returns GPS time tuple (GPSWeek, SOW, DOW, SOD)
                            (GPS week, Second of week, Day of week 0...6, Second of day))
    """

    if fromYearf:
        refday = TimefromYearf(
            refday,
        )

    refdayt = refday + datetime.timedelta(**shifTime(days))
    tmp = refdayt.timetuple()[0:6]

    if mday:
        return gpsFromUTC(
            *tmp[0:3], hour=12, min=0, sec=0, leapSecs=leapSecs, gpst=gpst
        )
    else:
        return gpsFromUTC(*tmp, leapSecs=leapSecs, gpst=gpst)


def gpsWeekDay(days=0, refday=currDate(), fromYearf=False):
    """
    Convenienc function
    Returns tuple gps Week and day of Week
    """
    return gpsfDateTime(days=0, refday=refday, fromYearf=False, mday=False)[0:3:2]


def datepathlist(
    stringformat, lfrequency, starttime=None, endtime=None, datelist=[], closed="left"
):
    """
    returns  a list of strings formated according to stringformat with
    frequency lfrequency for a given interval between starttime and endtime,
    can also accept a list of dates which will be converted according to
    stringformat. defaults to single element list with formated current date.

    input:
        stringformat: A String determinaning the output format of the input time
                      formated according to format codes that the C standard
                      (1989 version) requires, see documentation for datetime module
                      for details.                      Example, the date 2015-10-01
                      String = "/data/%Y/#b/VONC/15s_24hr/rinex/VONCR2inexO.Z" ->
                                /data/2015/oct/VONC/15s_24hr/rinex/VONC2740.15O.Z
                      Included are  some special case formating
                           #gpsw -> GPS week
                           #b -> %b converted to all lowercase.
                           #Rin2 -> %j(session).%y where session is a single character
                           reprecenting the session of the day. The coding depedepends
                           the (lfrequency).
                           #8hRin2 -> special case of 8h rinex files will overite lfrequency
                           by padding session parameter to {1, 2, 3}
                           #datelist -> returns a list of datetimeobjects instead of a string


        lfrequency:   A string defining the frequency of the datetime list created. uses
                      pandas.date_range to create the list (See pandas date_range function
                      for parameters but most common converion letters are
                      frequency letters, H -> hour, D -> day  A -> year
                      (and Y for newer versions of pandas)
                      precead  with a number to specify number of units.
                      examples. 3H -> 3 hours, 4D -> 4 days, 2A -> 2 years
                      The session parameter in stringformat are treated
                      differently depending lfrequency,
                                    lfrequency >= day -> session = 0
                                    lfrequency < day  -> session = {a,b,c ... x}
                                    lfrequency = 8H   -> session = {a, i, q}

        starttime:
                      datetime object reprecenting the start of the period
                      defaults to None, is set to datetime.datetime.utcnow()
                      if datelist is empty
        endtime:
                      datetime object reprecenting the end of the period
                      defaults to None , is set to datetime.datetime.utcnow()
                      if datelist is empty
        datelist:
                      Optional list of datetime object can be passed then
                      starttime and endtime are ignored.
        closed:
                      A string passed to pd.date_range, controls how interval
                      endpoints are treaded with given frequency "left", "right" or None,
                      See doc from pd.date_range
                      Defaults to "left"

    output:
                Returns list of strings with time codes formated according to input String.
                Example the date 2015-10-01:
                 stringformat -> "/data/%Y/#b/VONC/15s_24hr/rinex/VONCR2inexO.Z"
                 output       -> "/data/2015/oct/VONC/15s_24hr/rinex/VONC2740.15O.Z"

    """

    import re
    import datetime
    import pandas as pd

    today = datetime.datetime.utcnow()

    # special home made formating
    gpswmatch = re.compile(r"\w*(#gpsw)\w*").search(stringformat)  # use GPS week
    wrepl = ""
    rmatch = re.compile(r"\w*(#Rin2)\w*").search(
        stringformat
    )  # use GPS standard name RINEX2 name
    rrepl = ""
    r8hmatch = re.compile(r"\w*(#8hRin)\w*").search(
        stringformat
    )  # use GPS standard name RINEX2 name
    r8hrepl = ""
    bbbmatch = re.compile(r"\w*(#b)\w*").search(
        stringformat
    )  # use all lower case for 3 letter month
    bbbrepl = ""

    datelistmatch = re.compile(r"\w*(#datelist)\w*").search(
        stringformat
    )  # Return a list of datetime objects
    # -----------

    if (endtime is None) and not datelist:
        endtime = today
    elif (starttime is None) and not datelist:
        starttime = endtime = today
        datelist = [today]

    if datelist:
        pass
    elif lfrequency == "8H" or r8hmatch:
        mod = endtime - datetime.datetime.combine(endtime.date(), datetime.time(0))

        if mod > datetime.timedelta(16):
            mod += datetime.timedelta(16)
        elif mod > datetime.timedelta(8):
            mod += datetime.timedelta(8)

        if today - starttime > datetime.timedelta(hours=8):
            datelist = pd.date_range(
                starttime - mod, endtime - mod, freq=lfrequency, closed=closed
            ).tolist()
        else:
            datelist = [today - mod]

    else:
        hourshift = datetime.timedelta(hours=0)
        datelist = pd.date_range(
            starttime, endtime - hourshift, freq=lfrequency, inclusive=closed
        ).tolist()
        if not datelist:
            datelist = [endtime]

    if datelistmatch:
        return datelist

    stringlist = []
    for item in datelist:
        if rmatch or r8hmatch:  # form H or 8H rinex formating
            if rmatch:  # for rinex formating
                if lfrequency[-1] == "H":
                    hour = hourABC(item.hour)
                else:
                    hour = 0
            else:  # the specal case of 8H files
                hour = hour8hABC(item.hour)

            doy = int(item.strftime("%j"))
            yr = int(item.strftime("%y"))
            rrepl = "%.3d%s.%.2d" % (doy, hour, yr)

        if gpswmatch:  # for GPS week
            wrepl = "{0:04d}".format(gpsWeekDay(refday=item)[0])

        if bbbmatch:  # for lower case three letter month name Jan -> jan ...
            bbbrepl = "{:%b}".format(item).lower()

        # replacing special formating strings with the values
        pformat = re.sub("#gpsw", wrepl, stringformat)
        pformat = re.sub("#8hRin2", rrepl, pformat)
        pformat = re.sub("#Rin2", rrepl, pformat)
        pformat = re.sub("#b", bbbrepl, pformat)
        pformat = item.strftime(pformat)
        stringlist.append(pformat)

    return stringlist


############################################
# derived functions


def currTime(String):
    """
    Returns the current local time in a format determent by String

    input:
        String: A String determinaning the output format of the current time
                formated according to format codes that the C standard (1989 version) requires,
                see documentation for datetime module. Example
                Example,  String = "%Y %j %H:%M:%S %f" -> '2013 060 16:03:54 970424'
                See datetime documentation for details

    output:
                Returns the current time formated according to input String.

    """

    return datetime.datetime.now(tzlocal()).strftime(String)


def DayofYear(days=0, year=None, month=None, day=None):
    """
    Returns the day of year, "days" (defaults to 0) relative to the date given
    i.e. (year,month,day) (defaults to today)
    No argument returns the day of today

    input:
        days: Day relative to (year,month,day) or today if (year,month,day) not given
        year: Four digit year "yyyy". Example 2013
        month: Month in integer from 1-12
        day: Day of month as integer 1-(28-31) depending on month
    output:
        doy: Integer containing day of year. Exampls (2013,1,3) -> 60
                spans 1 -365(366 if leap year)
    """

    # if type(days) is int:
    #    tmp = {'days':days}
    #    days = tmp

    if year and month and day:
        nday = datetime.date(year, month, day) + datetime.timedelta(**shifTime(days))
        doy = nday.timetuple()[7]
    else:
        nday = datetime.date.today() + datetime.timedelta(**shifTime(days))
        doy = nday.timetuple()[7]

    return doy


def DaysinYear(year=None):
    """
    Returns the last day of year 365 or 366, (defaults to current year)

    input:
        year: Integer or floating point year (defaults to current year)
    out:
        daysinyear: Returns and integer value, the last day of the year  365 or 366
    """

    if year == None:  # defaults to current year
        year = datetime.date.today().year

    year = int(math.floor(year))  # allow for floating point year
    daysinyear = (
        366 if calendar.isleap(year) else 365
    )  # checking if it is leap year and assigning the correct day number

    return daysinyear


def yearDoy(yearf):
    """
    simple wrapper that calls TimefromYearf, to return a date in the form "year-doyT" from fractional year.
    convinient for fancy time labels in GMT hence the T.
    """
    return TimefromYearf(
        yearf,
        "%Y-%jT",
    )


def currYearfDate(days=0, refday=datetime.date.today(), fromYearf=True):
    """
    Wrapper for currDate() to return the date, "days" from "refday"
    in decimal year, defaults to current day
    """

    return currDate(days=days, refday=refday, String="yearf", fromYearf=fromYearf)


def currYear():
    """
    Current year in YYYY
    """
    return datetime.date.today().year


def shlyear(yyyy=currYear(), change=True):
    """
    Changes a year from two digit format to four and wize versa.
    input:
        YYYY: Year in YYYY or YY (defaults to current year)
        change: True of False convinies in case we want to pass YYYY unchanged through the function

    output: Year converterded from two->four or four->two digit form.
    returns current year in two digit form in the apsence of input
    """
    if len(str(abs(yyyy))) == 4 and change is True:
        yyyy = datetime.datetime.strptime(str(yyyy), "%Y")
        return yyyy.strftime("%y")
    elif len(str(abs(yyyy))) <= 2 and change is True:
        yyyy = datetime.datetime.strptime("%02d" % yyyy, "%y")
        return yyyy.strftime("%Y")
    elif change is False:
        return yyyy


def dateTuple(days=0, refday=datetime.datetime.today(), String=None, fromYearf=False):
    """
    Return tuple with different elements of a given date
    (year, month, day of month, day of year, fractional year, gps week, gps day of week)
    """

    # (Week,dow) = gpsWeekDay(days,refday,fromYearf)
    day = currDatetime(days, refday, String=String)
    month = day.strftime("%b")
    day = day.timetuple()
    return (
        day[0:3]
        + day[7:8]
        + (currYearfDate(days, refday),)
        + gpsWeekDay(days, refday, fromYearf)
        + (int(str(day[0])[-1]),)
        + (int(shlyear(day[0])),)
        + (month,)
    )


def hourABC(Hour=datetime.datetime.now().hour):
    """
    Return the hour as an alphabetica letter i.e. 00 -> a, 01 -> b ... 23 -> x
    """

    hourdict = dict(enumerate(string.ascii_lowercase, 0))

    return hourdict[Hour]


def ABChour(HourA):
    """
    Return the inverse of hourABC and hour8hABC
    """

    hourdict = dict(enumerate(string.ascii_lowercase, 0))
    if HourA == "0":
        return 0
    if HourA == "1":
        return 8
    if HourA == "2":
        return 16

    for key, value in hourdict.items():
        if value == HourA.lower():
            return key

    return ""


def hour8hABC(Hour=0):
    """
    returns hour 0, 8 and 16 as 0, 1 and 2
    IMO special case for 8hr rinex sessions.
    """

    hourdict = {
        0: 0,
        8: 1,
        16: 2,
    }

    return hourdict[Hour]


# Temporary functions to deal with numpy arrays Will become apsolete when implementd directly in the main moduvls


def convfromYearf(yearf, String=None):
    # from floating point year to floating point ordinal

    import numpy as np

    tmp = list(range(len(yearf)))

    for i in range(len(yearf)):
        if String:
            tmp[i] = TimefromYearf(yearf[i], String)
        else:
            tmp[i] = TimefromYearf(yearf[i])

    return np.asarray(tmp)


# functions using gps week and day of week ----------------


def datefRinex(rinex_list):
    """
    Return datetime object from rinex format
    """

    import os

    date_list = []

    for rinex in rinex_list:
        basename = os.path.basename(rinex)
        doy = basename[4:7]
        yy = basename[9:11]
        session = ABChour(basename[7:8])
        date_list.append(
            datetime.datetime.strptime(
                "{0}-{1}:{2:02d}".format(yy, doy, session), "%y-%j:%H"
            )
        )

    return date_list


def datefgpsWeekSOW(gpsWeek, SOW, String=None, leapSecs=None, mDay=False):
    """
    Return date (time) converted from GPS Week and Second of week (SOW)
    takes arguments:
        gpsWeek: An integer number of week since 1980-01-06 00:00:00

        SOW: Float Second of week (SOW) Then set

        String: output format See datetime for reference.
            None (Default), returns a python datetime object.
            For special formating:
            "yearf", will return date (time) in fractional year
            "tuple", will return a tuple with date (time)

        leapSecs: number of leap seconds to take into acount.


        mDay: Boolean Defaulsts to False returns date at 12 PM (noon),
               False return input time in second accuracy


    Returns date (time)

    """

    print("SOW: {}".format(SOW))
    print("gpsWeek: {}".format(gpsWeek))
    day = datetime.datetime(*UTCFromGps(gpsWeek, SOW, leapSecs=leapSecs))

    if mDay:
        day = day.replace(hour=12, minute=0, second=0)

    if String == "yearf":
        return TimetoYearf(*day.timetuple()[0:6])
    elif String == "tuple":
        return day.timetuple()[0:6]
    elif String:
        return day.strftime(String)
    else:
        return day


def datefgpsWeekDOW(gpsWeek, DOW, String=None, leapSecs=None, mDay=True):
    """
    Return date (time) converted from GPS Week and Day of week (DOW)
    takes arguments:

        DOW: integer Day of week

        See datefgpsWeekSOW for other arguments


    Returns date (time)

    """

    SOW = (DOW + 1) * secsInDay
    return datefgpsWeekSOW(gpsWeek, SOW, String=String, leapSecs=leapSecs, mDay=mDay)


def datefgpsWeekDOWSOD(gpsWeek, DOW, SOD, String=None, leapSecs=None, mDay=False):
    """
    Return date (time) converted from GPS Week and Day of week (DOW)
    takes arguments:

        DOW: integer Day of week
        SOD: float second of day

        See datefgpsWeekSOW for other arguments


    Returns date (time)

    """

    SOW = DOW * secsInDay + SOD
    return datefgpsWeekSOW(gpsWeek, SOW, String=String, leapSecs=leapSecs, mDay=mDay)


def datefgpsWeekDoy(gpsWeek, Doy, String=None, leapSecs=None):
    """
    Return date converted from GPS Week and Day of year
    """
    SOW = 1 * secsInDay
    day = datetime.datetime(*UTCFromGps(gpsWeek, SOW, leapSecs=leapSecs)[0:3])
    year0 = day.timetuple()[0]
    doy0 = day.timetuple()[7]

    daysinyear0 = DaysinYear(year0)
    daystoYend = daysinyear0 - doy0

    if doy0 <= Doy < doy0 + 7:  # check if doy is in the given week
        DOW = Doy - doy0
    elif (
        daystoYend < 6 and daysinyear0 + Doy - doy0 < 7
    ):  # in case it is the end of year
        DOW = daysinyear0 + Doy - doy0
    else:
        DOW = 0
        print(
            "ERROR: Doy %s is not in week %s returning date of day 0 of week %s"
            % (Doy, gpsWeek, gpsWeek)
        )

    day = day + datetime.timedelta(DOW)

    if String == "yearf":
        return TimetoYearf(*day.timetuple()[0:3])
    elif String == "tuple":
        return day.timetuple()[0:3]
    elif String:
        return day.strftime(String)
    else:
        return day


def toDatetime(dStr, fStr):
    """
    Convert date/time Strings to datetime objects accorting to formating rule defined in fStr

    input:

        dStr: (list of) String(s)  holding a date and/or time

        fStr: formating rule constituding the  following input formats
            default: fStr formated according to standard rules see for example datetime documentation for formating
            (i.e dStr=20150120 entailes fStr=%Y%m%d )

            yearf: decimal year
            w-dow: GPS week and day of week on the form WWWW-DOW (example 1820-3, where DOW is sunday = 0 ... 6 = saturday)
            w-dow-sod: GPS week and day of week on the form WWWW-DOW-SOD (example 1820-3-100, where DOW is sunday = 0 ... 6 = saturday)
            w-sow: GPS week and second of week on the form WWWW-SOW (example 1820-3000, where SOW is number of seconds since week started)
            w-dow-sod: GPS week - day of week - second of daym on the form WWWW-DOW-SOD (example 1820-1-18)
            w-doy: GPS week and day of year on the form WWWW-DOY
            Rinex: converts rinex format to rinex

    output:
        returns datetime object.

    """

    if type(dStr) == datetime.datetime:
        day = dStr

    elif fStr == "yearf":
        day = TimefromYearf(float(dStr))

    elif fStr == "w-dow":
        wdow = tuple([int(i) for i in dStr.split("-")])
        day = datefgpsWeekDOW(*wdow)

    elif fStr == "w-dow-sod":
        wdowsod = tuple([int(i) for i in dStr.split("-")])
        day = datefgpsWeekDOWSOD(*wdowsod)

    elif fStr == "w-sow":
        wsow = tuple([int(i) for i in dStr.split("-")])
        day = datefgpsWeekSOW(*wsow)

    elif fStr == "w-doy":
        wdoy = tuple([int(i) for i in dStr.split("-")])
        day = datefgpsWeekDoy(*wdoy)

    elif fStr == "Rinex":
        day = datefRinex(dstr)

    else:
        day = datetime.datetime.strptime(dStr, fStr)

    # returning datetime object
    return day


def toDatetimel(dStrlist, fStr):
    """
    A simple wrapper around toDatetime to allow for list input works like toDatetime if dStrlist is a single object.

    input:

        dStr: (list of) String(s)  holding a date and/or time

        fStr: See docstring of toDatetime

    output:
        returns a list of datetime objects.

    """

    # To allow for single object input as well, otherwise python will treat a string as a list in the for loop
    if type(dStrlist) is not list:
        dStrlist = [dStrlist]

    dStrlist = [
        toDatetime(dStr, fStr) for dStr in dStrlist
    ]  # converting to a list of datetime strings

    if len(dStrlist) == 1:  # toDatetimel can be replaced by toDatetime
        return dStrlist[0]
    else:
        return dStrlist


HOURS_PER_DAY = 24.0
MINUTES_PER_DAY = 60.0 * HOURS_PER_DAY
SECONDS_PER_DAY = 60.0 * MINUTES_PER_DAY
MUSECONDS_PER_DAY = 1e6 * SECONDS_PER_DAY
SEC_PER_MIN = 60
SEC_PER_HOUR = 3600
SEC_PER_DAY = SEC_PER_HOUR * 24
SEC_PER_WEEK = SEC_PER_DAY * 7


def _to_ordinalf(dt):
    """
    Convert :mod:`datetime` to the Gregorian date as UTC float days,
    preserving hours, minutes, seconds and microseconds.  Return value
    is a :func:`float`.
    """

    if hasattr(dt, "tzinfo") and dt.tzinfo is not None:
        delta = dt.tzinfo.utcoffset(dt)
        if delta is not None:
            dt -= delta

    base = float(dt.toordinal())
    if hasattr(dt, "hour"):
        base += (
            dt.hour / HOURS_PER_DAY
            + dt.minute / MINUTES_PER_DAY
            + dt.second / SECONDS_PER_DAY
            + dt.microsecond / MUSECONDS_PER_DAY
        )
    return base
